/*
OpenTherm.h - OpenTherm Library for the ESP8266/Arduino platform
https://github.com/ihormelnyk/OpenTherm
http://ihormelnyk.com/pages/OpenTherm
Licensed under MIT license
Copyright 2018, Ihor Melnyk

Frame Structure:
P MGS-TYPE SPARE DATA-ID  DATA-VALUE
0 000      0000  00000000 00000000 00000000
*/

#ifndef OpenTherm_h
#define OpenTherm_h

#include <stdint.h>
#include <Arduino.h>

enum OpenThermResponseStatus {
	NONE,
	SUCCESS,
	INVALID,
	TIMEOUT
};

enum OpenThermRequestType {
	READ,
	WRITE
};

enum OpenThermMessageID {
	Status, // flag8 / flag8 Флаги состояния Master и Slave.
	TSet, // f8.8 Контрольная уставка, т.е. уставка температуры воды в ЦТ (° C)
	MConfigMMemberIDcode, // flag8 / u8 Основные флаги конфигурации / главный код участника
	SConfigSMemberIDcode, // flag8 / u8 Флаги конфигурации ведомого / код ведомого участника
	Command, // u8 / u8 Удаленная команда
	ASFflags, // / OEM-код ошибки flag8 / u8 Флаги ошибок конкретного приложения и код ошибки OEM
	RBPflags, // flag8 / flag8 Удаленная передача параметров бойлера, флаги включения и чтения / записи
	CoolingControl, // f8.8 Сигнал управления охлаждением (%)
	TsetCH2, // f8.8 Контрольная уставка для цепи 2e CH (° C)
	TrOverride, // f8.8 Уставка комнаты удаленной переопределения
	TSP, // u8 / u8 Количество прозрачных-подчиненных параметров, поддерживаемых ведомым
	TSPindexTSPvalue, // u8 / u8 Индексный номер / Значение параметра прозрачного подчиненного, на который ссылаются.
	FHBsize, // u8 / u8 Размер буфера истории ошибок, поддерживаемого ведомым
	FHBindexFHBvalue, // u8 / u8 Индексный номер / Значение записи буфера , относящегося к истории ошибок.
	MaxRelModLevelSetting, // f8.8 Настройка максимального относительного уровня модуляции (%)
	MaxCapacityMinModLevel, // u8 / u8 Максимальная мощность котла (кВт) / Минимальный уровень модуляции котла (%)
	TrSet, // f8.8 Уставка комнаты (° C)
	RelModLevel, // f8.8 Относительный уровень модуляции (%)
	CHPressure, // f8.8 Давление воды в контуре ЦО (бар)
	DHWFlowRate, // f8.8 Расход воды в контуре ГВС. (л / мин)
	DayTime, // special / u8 День недели и время суток
	Date, // u8 / u8 Календарная дата
	Year, // u16 Календарный год
	TrSetCH2, // f8.8 Комнатное заданное значение для 2-го канала CH (° C)
	Tr, // f8.8 Комнатная температура (° C)
	Tboiler, // f8.8 Температура воды в котле (° C)
	Tdhw, // f8.8 температура ГВС (° C)
	Toutside, // f8.8 Наружная температура (° C)
	Tret, // f8.8 Температура возвратной воды (° C)
	Tstorage, // ф8.8 Солнечная температура хранения (° С)
	Tcollector, // f8.8 Температура солнечного коллектора (° C)
	TflowCH2, // f8.8 Температура проточной воды в контуре CH2 (° C)
	Tdhw2, // f8.8 Температура горячей воды для бытового потребления 2 (° C)
	Texhaust, // s16 Температура выхлопных газов котла (° C)
	TdhwSetUBTdhwSetLB = 48 , // s8 / s8 Верхняя и нижняя границы уставки ГВС для регулировки (° C)
	MaxTSetUBMaxTSetLB, // s8 / s8 Макс. CH Уставка воды верхняя и нижняя границы для регулировки (° C)
	HcratioUBHcratioLB, // s8 / s8 OTC коэффициент тепловой кривой верхняя и нижняя границы для регулировки  
	TdhwSet = 56 , // f8.8 Уставка ГВС (° C) (дистанционный параметр 1)
	MaxTSet, // f8.8 Максимальное заданное значение температуры воды в канале (° C) (дистанционные параметры 2)
	Hcratio, // f8.8 Отношение тепловой кривой OTC (° C) (Удаленный параметр 3)
	RemoteOverrideFunction = 100 , // flag8 / - Функция ручных и программных изменений в основной и удаленной уставках комнаты.
	OEMDiagnosticCode = 115 , // u16 Специфичный для OEM диагностический / сервисный код
	BurnerStarts, // u16 Количество запусков горелки
	CHPumpStarts, // u16 Количество пусков насоса ЦО
	DHWPumpValveStarts, // u16 Количество пусков ГВС / клапан
	DHWBurnerStarts, // u16 Количество пусковых горелок в режиме ГВС
	BurnerOperationHours, // u16 Количество часов, в течение которых работает горелка (т. Е. Включенное пламя)
	CHPumpOperationHours, // u16 Количество часов, в течение которых работал насос CH
	DHWPumpValveOperationHours, // u16 Количество часов, в течение которых насос ГВС работал или клапан ГВС был открыт
	DHWBurnerOperationHours, // u16 Количество часов работы горелки в режиме ГВС
	OpenThermVersionMaster, // f8.8 Реализованная версия спецификации протокола OpenTherm в мастере.
	OpenThermVersionSlave, // f8.8 Реализованная версия спецификации протокола OpenTherm в подчиненном устройстве.
	MasterVersion, // u8 / u8 Номер и тип версии основного продукта
	SlaveVersion, // u8 / u8 Номер и тип версии подчиненного продукта
};

enum OpenThermStatus {
	NOT_INITIALIZED,
	READY,
	DELAY,
	REQUEST_SENDING,
	RESPONSE_WAITING,
	RESPONSE_START_BIT,
	RESPONSE_RECEIVING,
	RESPONSE_READY,
	RESPONSE_INVALID	
};

class OpenTherm
{
private:
	const int inPin;
	const int outPin;	

	volatile OpenThermStatus status;
	volatile unsigned long response;
	volatile OpenThermResponseStatus responseStatus;
	volatile unsigned long responseTimestamp;
	volatile byte responseBitIndex;
	
	int readState();
	void setActiveState();
	void setIdleState();	
	void activateBoiler();

	void sendBit(bool high);
	bool parity(unsigned long frame);
	bool isValidResponse(unsigned long response);
	void(*handleInterruptCallback)();
	void(*processResponseCallback)(unsigned long, OpenThermResponseStatus);
public:	
	OpenTherm(int inPin = 4, int outPin = 5);
	void begin(void(*handleInterruptCallback)(void));
	void begin(void(*handleInterruptCallback)(void), void(*processResponseCallback)(unsigned long, OpenThermResponseStatus));
	bool isReady();
	unsigned long sendRequest(unsigned long request);
	bool sendRequestAync(unsigned long request);
	unsigned long buildRequest(OpenThermRequestType type, OpenThermMessageID id, unsigned int data);
	OpenThermResponseStatus getLastResponseStatus();
	void handleInterrupt();	
	void process();
	void end();

	//building requests // создание запросов
	unsigned long buildSetBoilerStatusRequest(bool enableCentralHeating, bool enableHotWater = false, bool enableCooling = false, bool enableOutsideTemperatureCompensation = false, bool enableCentralHeating2 = false);
	unsigned long buildSetBoilerTemperatureRequest(float temperature);
	unsigned long buildGetBoilerTemperatureRequest();
	unsigned long buildGetModLevelRequest();

	//parsing responses // парсинг ответов
	bool isFault(unsigned long response);
	bool isCentralHeatingEnabled(unsigned long response);
	bool isHotWaterEnabled(unsigned long response);
	bool isFlameOn(unsigned long response);
	bool isCoolingEnabled(unsigned long response);
	bool isDiagnostic(unsigned long response);
	uint16_t getUInt(const unsigned long response) const;
	float getFloat(const unsigned long response) const;
	float getTemperature(unsigned long response);
	float getlevel(unsigned long response);
	unsigned int temperatureToData(float temperature);

	//basic requests	// основные запросы
	unsigned long setBoilerStatus(bool enableCentralHeating, bool enableHotWater = false, bool enableCooling = false, bool enableOutsideTemperatureCompensation = false, bool enableCentralHeating2 = false);	
	bool setBoilerTemperature(float temperature);
	float getBoilerTemperature();
	float getModLevel();
};

#endif // OpenTherm_h
